'use strict';

var R = require('ramda');
var util = require('util');
var rescapeRamda = require('rescape-ramda');
var requestHelpers = require('./requestHelpers.js');
var AC = require('@apollo/client');
var reactApollo = require('react-apollo');
var rescapeHelpersComponent = require('rescape-helpers-component');
var graphql = require('graphql');
var rescapeLog = require('rescape-log');
var apolloClient = require('./apolloClient.js');
var containerHelpers = require('./containerHelpers.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var R__default = /*#__PURE__*/_interopDefaultLegacy(R);
var AC__default = /*#__PURE__*/_interopDefaultLegacy(AC);

/**
 * Created by Andy Likuski on 2020.05.15
 * Copyright (c) 2020 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const nameComponent = (name, component) => {
  component.displayName = name;
  return component;
};
/**
 * For some reason Query and Mutation don't have displayNames, so use their type.name
 * @param component
 * @return {*}
 */

const getDisplayName = component => {
  return rescapeRamda.strPathOr(null, 'displayName', component) || rescapeRamda.strPathOr('Undefined', 'type.name', component);
};
/**
 * Given a child component and its parent component, returns a function that expects the children
 * render prop for the childComponent (i.e. the render prop that makes grandchildren). Providing the grandchildren
 * render prop results in calling the parentComponent with the childComponent, and the childComponent's
 * input props are modified to include the grandchildren render prop at props.children
 * @param childComponent
 * @param parentComponent
 * @return {function(*=): *}
 */


const embedComponents = (config, childComponent, parentComponent) => {
  // Wrap parentComponent in an HOC component that expects a render function, grandchildren
  const displayName = `${getDisplayName(parentComponent)}(${getDisplayName(childComponent)})`;
  return nameComponent(displayName, grandchildren => {
    // Only for tests that aren't using React
    if (R__default['default'].propOr(false, '_noReact', config)) {
      return parentComponent(nameComponent(displayName, props => {
        return childComponent(R__default['default'].merge(props, {
          children: grandchildren,
          render: grandchildren
        }));
      }));
    }

    return nameComponent(displayName, props => {
      // Pass a modified version of childComponent that adds a children render prop
      return parentComponent(nameComponent(displayName, p => {
        return childComponent(R__default['default'].merge(p, {
          children: props.children,
          render: props.children
        }));
      }));
    })({
      children: grandchildren
    });
  });
};
/**
 * Returns the props that are possible render props
 * @param props
 * @returns {*}
 */

const pickRenderProps = props => {
  return R__default['default'].pick(['render', 'children'], props);
};
const getRenderProp = props => {
  return R__default['default'].find(prop => R__default['default'].propOr(null, prop, props), ['render', 'children']);
};
const getRenderPropFunction = props => {
  return R__default['default'].prop(getRenderProp(props), props);
};
/**
 * If we have a component, wrap it in a Maybe so we can chain with the props that the component returns.
 * If it's a task chain it with the next task
 * @param {[Task|Object]} list Each is a task that resolves to the Apollo request response or returns an
 * Apollo Component
 * @param {[Task|Object]} componentOrProps For tasks this is props. For components, this is the child component.
 * When composed it returns the composed component expecting props.
 * Apollo Component
 * @return {Task|Object]} The resolution of the final task of list (the first in the list since we're
 * evaluating bottom to top) or returns the built up component from bottom to top, where the outermost component
 * is the from the bottom of list and the most embedded component is from the top of list. If a task
 * the return value will expect props to start the task evaluation chain. If a component, it will expect
 * to be given a component that is to be the child component of the innermost component (the one from the
 * top of list). The result of passing the child component is a composed component expecting props.
 */

const composeWithComponentMaybeOrTaskChain = list => {
  return props => {
    const renderProp = getRenderProp(props); // Delay the last item of the list (first evaluated) so we can give it its render function
    // to render its child. For subsequent items in the list, this is accomplished by embedComponents below
    // TODO we do all this so we can pass the render function from the top component through the intermediate
    // down to the lowest component. However it might be better if we can just give the render function to
    // the lowest component when we create it.

    const delayedList = R__default['default'].when(() => renderProp, list => R__default['default'].over(R__default['default'].lensIndex(-1), f => {
      return props => {
        // Delay evaluation by wrapping in a function expecting the children component
        // so we can link the first called component in list (the last one) to the second (the penultimate)
        const compFunc = children => {
          return f(R__default['default'].merge(props, {
            [renderProp]: children
          }));
        };

        compFunc.displayName = f.displayName;
        return compFunc;
      };
    }, list))(list);
    const composed = R__default['default'].composeWith((nextPropsToTaskOrComponent, lastTaskOrComponent) => {
      return R__default['default'].ifElse(res => R__default['default'].both(R__default['default'].is(Object), res => 'run' in res)(res), // Chain task with the next function
      // This gives the props that result from the task to a function that produces a task
      task => R__default['default'].chain(nextPropsToTaskOrComponent, task), // Pass the component to the next function, which produces the child component.
      // embedComponents makes an HOC from component and child component (nextPropsToTaskOrComponent)
      // so that component can pass props to the child component, including a children render prop
      // to the child component (the render prop to create the grandchildren).
      component => {
        return embedComponents(R__default['default'].pick(['_noReact'], props), nextPropsToTaskOrComponent, component);
      })(lastTaskOrComponent);
    })(delayedList); // For tasks, componentOrPropsOrBoth is always just props
    // For components, componentOrPropsOrBoth is the child component. Since we must pass composed the props
    // first, we pass props below and then component after.
    // We could optional expect props with/ a children props as the child component here,
    // but I think we'll always pass the child component before the props

    const taskOrComponent = composed(props);
    return R__default['default'].ifElse(() => R__default['default'].hasIn('run', taskOrComponent), () => {
      // For tasks, just pass the props. This returns a task that is ready to execute
      if (!R__default['default'].hasIn('run', taskOrComponent)) {
        throw new Error('Task expected, but composed did not produce a task. Is this supposed to be a component and you forgot to pass a render function?');
      }

      return taskOrComponent;
    }, // Match the form of HOC(component)(props), even though composed expects props first
    props => {
      if (!R__default['default'].any(prop => R__default['default'].propOr(false, prop, props), ['render', 'children'])) {
        throw new Error(`Expected a render function in props but did't find one ${util.inspect(props)}`);
      } // For components, pass props, this produces a function that must be called
      // to create the correct chaining process between components
      // Pass the render prop. This passes the render prop from outermost component to innermost


      return taskOrComponent(R__default['default'].prop(renderProp, props));
    })(props);
  };
};

/**
 * Created by Andy Likuski on 2019.02.01
 * Copyright (c) 2019 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const {
  MissingFieldError
} = AC__default['default'];
const log = rescapeLog.loggers.get('rescapeDefault');
/**
 * Reads values loaded from the server that we know are now in the cache
 * @param {Object} apolloConfig
 * @param {Object} apolloConfig.apolloClient The authenticated Apollo Client
 * @param {Object} apolloConfig.options
 * @param {Function} apolloConfig.options.variables Optional filter of the props
 * @param {Object} options Query options for the Apollo Client See Apollo's Client.query docs
 * The main arguments for options are QueryOptions with query and variables. Example
 * query: gql`
 query regions($key: String!) {
          regions(key: $key) {
              id
              key
              name
          }
    }`,
 variables: {key: "earth"}
 * @returns {Object} The cache result
 */

const authApolloClientQueryCacheContainer = R__default['default'].curry((apolloConfig, options, props) => {
  const winnowedProps = requestHelpers._winnowRequestProps(apolloConfig, props); // readQuery isn't a promise, just a direct call I guess


  log.debug(`Query cache: ${graphql.print(options.query)} props: ${util.inspect(winnowedProps)}`);

  try {
    return {
      data: rescapeRamda.reqStrPathThrowing('apolloClient', apolloConfig).readQuery({
        variables: winnowedProps,
        ...options
      })
    };
  } catch (e) {
    if (!R__default['default'].is(MissingFieldError, e)) {
      throw e;
    }

    return {
      data: null
    };
  }
});
/**
 * Direct read from the cache
 */

const authApolloClientOrComponentQueryCacheContainer = R__default['default'].curry((apolloConfig, query, props) => {
  return R__default['default'].cond([// Apollo Client instance
  [R__default['default'].has('apolloClient'), apolloConfig => {
    return authApolloClientQueryCacheContainer(apolloConfig, query, props);
  }], // Apollo component instance
  [R__default['default'].T, // Since we aren't using a Query component, use an ApolloConsumer to get access to the
  // apollo client from the react context
  apolloConfig => {
    return rescapeHelpersComponent.e(reactApollo.ApolloConsumer, {}, apolloClient => {
      return containerHelpers.containerForApolloType(apolloClient, {
        render: getRenderPropFunction(props),
        response: authApolloClientQueryCacheContainer(R__default['default'].merge(apolloConfig, {
          apolloClient
        }), query, props)
      });
    });
  }]])(apolloConfig);
});
/**
 * Direct read fragment from the cache
 * @param {Object} apolloConfig
 * @param {Object} fragment The fragment query
 * @param {Object} props Not used in the fragment, but passed to the render prop for components
 * @param {Number|String} id The id for the fragment query
 */

const authApolloClientOrComponentReadFragmentCacheContainer = R__default['default'].curry((apolloConfig, {
  fragment
}, props, id) => {
  return R__default['default'].cond([// Apollo Client instance
  [R__default['default'].has('apolloClient'), apolloConfig => apolloClient.apolloClientReadFragmentCacheContainer(apolloConfig, fragment, id)], // Apollo component instance
  [R__default['default'].T, // Since we aren't using a Query component, use an ApolloConsumer to get access to the
  // apollo client from the react context
  apolloConfig => {
    return rescapeHelpersComponent.e(reactApollo.ApolloConsumer, {}, apolloClient$1 => {
      return containerHelpers.containerForApolloType(apolloConfig, {
        render: getRenderPropFunction(props),
        response: apolloClient.apolloClientReadFragmentCacheContainer(R__default['default'].merge(apolloConfig, {
          apolloClient: apolloClient$1
        }), fragment, id)
      });
    });
  }]])(apolloConfig);
});

exports.authApolloClientOrComponentQueryCacheContainer = authApolloClientOrComponentQueryCacheContainer;
exports.authApolloClientOrComponentReadFragmentCacheContainer = authApolloClientOrComponentReadFragmentCacheContainer;
exports.authApolloClientQueryCacheContainer = authApolloClientQueryCacheContainer;
exports.composeWithComponentMaybeOrTaskChain = composeWithComponentMaybeOrTaskChain;
exports.getRenderProp = getRenderProp;
exports.getRenderPropFunction = getRenderPropFunction;
exports.nameComponent = nameComponent;
exports.pickRenderProps = pickRenderProps;
//# sourceMappingURL=apolloClientCache-855e9291.js.map
