'use strict';

var R = require('ramda');
var rescapeRamda = require('rescape-ramda');
var requestHelpers = require('./requestHelpers.js');
var AC = require('@apollo/client');
var T = require('folktale/concurrency/task');
var rescapeValidate = require('rescape-validate');
var PropTypes = require('prop-types');
var mutationHelpers = require('./mutationHelpers.js');
var containerHelpers = require('./containerHelpers.js');
var apolloClientCache = require('./apolloClientCache-855e9291.js');
var queryHelpers = require('./queryHelpers.js');
var mutationCacheHelpers = require('./mutationCacheHelpers.js');
var queryCacheHelpers = require('./queryCacheHelpers.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var R__default = /*#__PURE__*/_interopDefaultLegacy(R);
var AC__default = /*#__PURE__*/_interopDefaultLegacy(AC);
var T__default = /*#__PURE__*/_interopDefaultLegacy(T);
var PropTypes__default = /*#__PURE__*/_interopDefaultLegacy(PropTypes);

/**
 * Created by Andy Likuski on 2020.08.18
 * Copyright (c) 2020 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const {
  MissingFieldError
} = AC__default['default'];
const {
  of
} = T__default['default'];
const userOutputParams = {
  id: 1,
  lastLogin: 1,
  username: 1,
  firstName: 1,
  lastName: 1,
  email: 1,
  isStaff: 1,
  isActive: 1,
  dateJoined: 1,
  ...requestHelpers.versionOutputParamsMixin
}; // Every complex input type needs a type specified in graphql. Our type names are
// always in the form [GrapheneFieldType]of[GrapheneModeType]RelatedReadInputType
// Following this location.data is represented as follows:
// TODO These value should be dervived from the schema

const userReadInputTypeMapper = {
  'data': 'DataTypeofUserTypeRelatedReadInputType'
};
/**
 * Reads the cache to check if the client is authenticated
 * It's technically possible to used an unauthed apolloClient to read from the cache
 * when the user is present, but hopefully this won't happen. It would be better
 * to check the apolloClient for th AuthLink header auth token, but I don't know how
 * to access the authLink since it's composed with other links. We could also but
 * an auth flag on the apolloClient object somewhere
 * @param {Object} apolloConfig
 * @param {Object} apolloConfig.apolloClient The Apollo client
 * @returns {Boolean} true if authenticated or false
 */

const isAuthenticatedLocal = apolloConfig => {
  // Unfortunately a cache miss throws
  try {
    return !!rescapeRamda.strPathOr(false, 'data.currentUser', queryCacheHelpers.makeQueryFromCacheContainer(apolloConfig, {
      name: 'currentUser',
      readInputTypeMapper: userReadInputTypeMapper,
      outputParams: userOutputParams
    }, {}));
  } catch (e) {
    if (R__default['default'].is(MissingFieldError)) {
      return false;
    }

    throw e;
  }
};
/**
 * Like isAuthenticatedLocal, but matches the style of asynchronous requests,
 * returning a task or apollo component
 *
 * @param {Object} apolloConfig
 * @param apolloConfig.apolloClient. If non-null then a task is returned. If null
 * an apollo component is returned
 * @param {Object} props
 * @returns {Task|Object} The authenticated user as a task or apollo component
 */

const authenticatedUserLocalContainer = (apolloConfig, props) => {
  // Unfortunately a cache miss throws
  try {
    return R__default['default'].compose( // Wrap in a task when we are doing apolloClient queries, otherwise we already have
    // a proper apollo container
    containerOrValue => R__default['default'].when(() => R__default['default'].propOr(false, 'apolloClient', apolloConfig), of)(containerOrValue), props => queryCacheHelpers.makeQueryFromCacheContainer(R__default['default'].merge(apolloConfig, {
      options: {
        variables: () => {
          return {};
        },
        // Pass through error so we can handle it in the component
        errorPolicy: 'all',
        partialRefetch: true
      }
    }), {
      name: 'currentUser',
      readInputTypeMapper: userReadInputTypeMapper,
      outputParams: userOutputParams
    }, props))(props);
  } catch (e) {
    if (R__default['default'].is(MissingFieldError, e)) {
      return containerHelpers.containerForApolloType(apolloConfig, {
        render: apolloClientCache.getRenderPropFunction(props),
        response: null
      });
    }

    throw e;
  }
};
/**
 * Queries users
 * @params {Object} apolloClient The Apollo Client
 * @params {Object} ouptputParams OutputParams for the query such as userOutputParams
 * @params {Object} props Unused but here to match the Apollo Component pattern. Use null or {}.
 * @returns {Task<Result>} A Task containing the Result.Ok with a User in an object with Result.Ok({data: currentUser: {}})
 * or errors in Result.Error({errors: [...]})
 */

const currentUserQueryContainer = rescapeValidate.v(R__default['default'].curry((apolloConfig, outputParams, props) => {
  return queryHelpers.makeQueryContainer(R__default['default'].merge(apolloConfig, {
    options: {
      variables: props => {
        // No arguments, the server resolves the current user based on authentication
        return {};
      },
      errorPolicy: 'all',
      partialRefetch: true
    }
  }), {
    // If we have to query for users separately use the limited output userStateOutputParamsCreator
    name: 'currentUser',
    readInputTypeMapper: userReadInputTypeMapper,
    outputParams
  }, props);
}), [['apolloConfig', PropTypes__default['default'].shape().isRequired], ['outputParams', PropTypes__default['default'].shape().isRequired], ['props', PropTypes__default['default'].shape()]], 'currentUserQueryContainer');

/**
 * Created by Andy Likuski on 2019.01.07
 * Copyright (c) 2019 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/***
 * The private test config is used for testing.
 * In production the application seeds the ApolloClient with settings from the application.
 * The production application would read a settings key from an environment variable and
 * pull the correct settings from a database or private file.
 * Settings are used to determine the shape and initial values of the Apollo in memory cache.
 * Since we have a settings class in the API, the settings must match the shape of the API's settings.data object,
 * unless they are cache only values like testAuthorization and mapboxApiAccessToken below.
 * See settingsStore.js
 */

var settings = {
  key: 'default',
  data: {
    domain: 'localhost',
    api: {
      protocol: 'http',
      host: 'localhost',
      port: '8008',
      path: '/graphql/'
    },
    // Used to authenticate with the API above in tests
    // @client only
    testAuthorization: {
      username: 'test',
      password: 'testpass'
    },
    // Overpass API configuration to play nice with the server's strict throttling
    overpass: {
      cellSize: 100,
      sleepBetweenCalls: 1000
    },
    mapbox: {
      // @client only
      mapboxAuthentication: {
        mapboxApiAccessToken: rescapeRamda.reqStrPathThrowing('MAPBOX_API_ACCESS_TOKEN', process.env)
      },
      // Initial viewport
      viewport: {
        zoom: 0,
        latitude: 0,
        longitude: 0
      }
    }
  }
};

const {
  of: of$1
} = T__default['default'];
/**
 * Created by Andy Likuski on 2020.03.20
 * Copyright (c) 2020 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
// Every complex input type needs a type specified in graphql. Our type names are
// always in the form [GrapheneFieldType]of[GrapheneModeType]RelatedReadInputType
// Following this location.data is represented as follows:
// TODO These value should be derived from the schema

const readInputTypeMapper = {};
/**
 * Uses the given configuration to extract the props that only belong in the cache
 * @param cacheOnlyObjs
 * @param cacheIdProps
 * @param props
 * @return {Object} Modified props
 */

const createCacheOnlyPropsForSettings = ({
  cacheOnlyObjs,
  cacheIdProps
}, props) => {
  return mutationCacheHelpers.createCacheOnlyProps({
    name: 'settings',
    cacheIdProps,
    cacheOnlyObjs
  }, props);
};
/**
 * Queries settings
 * @params {Object} apolloConfig The Apollo config. See makeQueryContainer for options
 * @params {Array|Object} outputParams OutputParams for the query such as defaultSettingsOutputParams
 * @params {Object} props Arguments for the Settings query. This can be {} or null to not filter.
 * @returns {Task} A Task containing the Settingss in an object with obj.data.settings or errors in obj.errors
 */

const makeSettingsQueryContainer = rescapeValidate.v(R__default['default'].curry((apolloConfig, {
  outputParams
}, props) => {
  return queryHelpers.makeQueryContainer(apolloConfig, {
    name: 'settings',
    readInputTypeMapper,
    outputParams
  }, props);
}), [['apolloConfig', PropTypes__default['default'].shape({
  apolloClient: PropTypes__default['default'].shape()
}).isRequired], ['queryStructure', PropTypes__default['default'].shape({
  outputParams: PropTypes__default['default'].oneOfType([PropTypes__default['default'].array, PropTypes__default['default'].shape()]).isRequired
})], ['props', PropTypes__default['default'].shape().isRequired]], 'makeSettingsQueryContainer');
/**
 * Makes a Settings mutation
 * @param {Object} apolloConfig Configuration of the Apollo Client when using one inst
 * @param {Object} apolloConfig.apolloClient An authorized Apollo Client
 * @param [String|Object] defaultSettingsOutputParams output parameters for the query in this style json format:
 *  ['id',
 *   {
 *        data: [
 *         'foo',
 *         {
 *            properties: [
 *             'type',
 *            ]
 *         },
 *         'bar',
 *       ]
 *    }
 *  ]
 *  @param {Object} props Object matching the shape of a settings. E.g. {id: 1, city: "Stavanger", data: {foo: 2}}
 *  @returns {Task|Just} A container. For ApolloClient mutations we get a Task back. For Apollo components
 *  we get a Just.Maybe back. In the future the latter will be a Task when Apollo and React enables async components
 */

const makeSettingsMutationContainer = rescapeValidate.v(R__default['default'].curry((apolloConfig, {
  cacheOnlyObjs,
  cacheIdProps,
  outputParams
}, props) => {
  return mutationHelpers.makeMutationRequestContainer(R__default['default'].merge(apolloConfig, {
    options: {
      update: (store, response) => {
        // Add mutate to response.data so we dont' have to guess if it's a create or update
        const settings = rescapeRamda.reqStrPathThrowing('data.mutate.settings', mutationHelpers.addMutateKeyToMutationResponse({
          silent: true
        }, response));
        makeSettingsCacheMutation(apolloConfig, {
          outputParams
        }, props, settings);
      }
    }
  }), {
    name: 'settings',
    outputParams
  }, // Remove client-side only values
  rescapeRamda.omitDeepPaths(cacheOnlyObjs, props));
}), [['apolloConfig', PropTypes__default['default'].shape().isRequired], ['mutationStructure', PropTypes__default['default'].shape({
  cacheOnlyObjs: PropTypes__default['default'].array.isRequired,
  cacheIdProps: PropTypes__default['default'].array.isRequired,
  outputParams: PropTypes__default['default'].oneOfType([PropTypes__default['default'].array, PropTypes__default['default'].shape()]).isRequired
}).isRequired], ['props', PropTypes__default['default'].shape().isRequired]], 'makeSettingsMutationContainer');
const makeSettingsCacheMutation = (apolloConfig, {
  outputParams
}, props, settings) => {
  // Add the cache only values to the persisted settings
  const propsWithCacheOnlyItems = mutationCacheHelpers.mergeCacheable({}, settings, props); // Mutate the cache to save settings to the database that are not stored on the server

  mutationCacheHelpers.makeCacheMutation(apolloConfig, {
    name: 'settings',
    // Use key instead of id in the case of the unauthenticated user needs to cache default settings
    idField: props => R__default['default'].propOr(R__default['default'].prop('key'), 'id', props),
    // output for the read fragment
    outputParams
  }, propsWithCacheOnlyItems);
};
/**
 * Writes or rewrites the default settings to the cache if needed.
 * Non-server values in the config are ignored and must be added manually.
 * If the apollo client isn't authenticated we write straight to cache
 * @param {Object} config
 * @param {Object} config.settings The settings to write. It must match Settings object of the Apollo schema,
 * although cache-only values can be included
 * @param {Object} config.defaultSettingsTypenames Typenmaes of the settings in the form
 * {
 *   __typename: string,
 *   data: {
 *     __typename: string
 *     foo: {...}
 *   }
 * }
 * Only need to write settings to the cache for an unauthed user when no settings are on the server (rare)
 */

const writeConfigToServerAndCache = config => {
  return (apolloClient, {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  }) => {
    // Only the settings are written to the server
    const configuredSettings = R__default['default'].prop('settings', config);
    const defaultSettingsTypenames = rescapeRamda.reqStrPathThrowing('settingsConfig.defaultSettingsTypenames', config);
    return rescapeRamda.composeWithChain([// Update/Create the default settings to the database. This puts them in the cache
    rescapeRamda.mapToNamedPathAndInputs('settingsWithoutCacheValues', 'data.mutate.settings', ({
      props,
      apolloConfig,
      settingsFromServer
    }) => {
      const settings = rescapeRamda.strPathOr({}, 'data.settings.0', settingsFromServer);
      return R__default['default'].ifElse(() => {
        // If we are authenticated and the server settings don't match the config, update
        // TODO This should only be done by admins
        return isAuthenticatedLocal(apolloConfig) && R__default['default'].not(R__default['default'].equals(settings, rescapeRamda.omitDeepPaths(cacheOnlyObjs, props)));
      }, () => {
        return makeSettingsMutationContainer(apolloConfig, {
          cacheOnlyObjs,
          cacheIdProps,
          outputParams: settingsOutputParams
        }, R__default['default'].merge(props, R__default['default'].pick(['id'], settings)));
      }, () => {
        // Write the server or configured values to the cache manually
        // If we have settings from the server, they will already be in the cache,
        // but we need to write any non server settings
        // If we don't have settings from the server and we aren't authenticated, just
        // cache the configured settings
        const settingsToCache = R__default['default'].length(R__default['default'].keys(settings)) ? settings : R__default['default'].mergeDeepRight(configuredSettings, // TODO this should come from the remote schema so it can be customized
        // to the app's settings
        defaultSettingsTypenames);
        makeSettingsCacheMutation(apolloConfig, {
          outputParams: settingsOutputParams
        }, configuredSettings, settingsToCache);
        return of$1({
          data: {
            mutate: {
              settings: settingsToCache
            }
          }
        });
      })();
    }), // Fetch the props if they exist on the server
    rescapeRamda.mapToNamedResponseAndInputs('settingsFromServer', ({
      apolloConfig,
      props
    }) => {
      return makeSettingsQueryContainer(R__default['default'].merge(apolloConfig, {
        options: {
          skip: localStorage.getItem('token') !== 'null',
          fetchPolicy: 'network-only'
        }
      }), {
        outputParams: requestHelpers.omitClientFields(settingsOutputParams)
      }, R__default['default'].pick(['key'], props));
    })])({
      apolloConfig: {
        apolloClient
      },
      props: configuredSettings
    });
  };
};

/**
 * Created by Andy Likuski on 2018.12.31
 * Copyright (c) 2018 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

const defaultSettingsTypenames = {
  __typename: 'SettingsType',
  data: {
    __typename: 'SettingsDataType',
    api: {
      _typename: 'SettingsApiDataType'
    },
    overpass: {
      _typename: 'SettingsOverpassDataType'
    },
    mapbox: {
      _typename: 'MapboxApiDataType'
    }
  }
}; // Global settings.
// omitCacheOnlyFields to true to omit cache only fields from the query

const defaultSettingsOutputParams = {
  id: 1,
  key: 1,
  data: {
    domain: 1,
    api: {
      protocol: 1,
      host: 1,
      port: 1,
      path: 1
    },
    'testAuthorization @client': {
      username: 1,
      password: 1
    },
    // Overpass API configuration to play nice with the server's strict throttling
    overpass: {
      cellSize: 1,
      sleepBetweenCalls: 1
    },
    mapbox: {
      'mapboxAuthentication @client': {
        mapboxApiAccessToken: 1
      },
      viewport: {
        zoom: 1,
        latitude: 1,
        longitude: 1
      }
    }
  }
}; // Paths to prop values that we don't store in the database, but only in the cache
// The prop paths are marked with a client directive when querying (see defaultSettingsOutputParams)
// so we never try to load them from the database.

const defaultSettingsCacheOnlyObjs = ['data.testAuthorization', 'data.mapbox.mapboxAuthentication']; // These values come back from the server and get merged into cacheOnlyProps for identification

const defaultSettingsCacheIdProps = ['id', '__typename', 'data.__typename', 'data.mapbox.__typename'];
/**
 * Writes or rewrites the default settings to the cache
 */

const writeDefaultSettingsToCache = writeConfigToServerAndCache({
  settings,
  settingsConfig: {
    defaultSettingsTypenames
  }
});

exports.authenticatedUserLocalContainer = authenticatedUserLocalContainer;
exports.createCacheOnlyPropsForSettings = createCacheOnlyPropsForSettings;
exports.currentUserQueryContainer = currentUserQueryContainer;
exports.defaultSettingsCacheIdProps = defaultSettingsCacheIdProps;
exports.defaultSettingsCacheOnlyObjs = defaultSettingsCacheOnlyObjs;
exports.defaultSettingsOutputParams = defaultSettingsOutputParams;
exports.defaultSettingsTypenames = defaultSettingsTypenames;
exports.isAuthenticatedLocal = isAuthenticatedLocal;
exports.makeSettingsMutationContainer = makeSettingsMutationContainer;
exports.makeSettingsQueryContainer = makeSettingsQueryContainer;
exports.readInputTypeMapper = readInputTypeMapper;
exports.settings = settings;
exports.userOutputParams = userOutputParams;
exports.userReadInputTypeMapper = userReadInputTypeMapper;
exports.writeConfigToServerAndCache = writeConfigToServerAndCache;
exports.writeDefaultSettingsToCache = writeDefaultSettingsToCache;
//# sourceMappingURL=defaultSettingsStore-b9a20f38.js.map
