'use strict';

var R = require('ramda');
var rescapeRamda = require('rescape-ramda');
var AC = require('@apollo/client');
var T = require('folktale/concurrency/task');
var rescapeLog = require('rescape-log');
var apolloClient = require('./apolloClient.js');
var mutationHelpers = require('./mutationHelpers.js');
var containerHelpers = require('./containerHelpers.js');
var apolloClientCache = require('./apolloClientCache-855e9291.js');
var mutationCacheHelpers = require('./mutationCacheHelpers.js');
var queryCacheHelpers = require('./queryCacheHelpers.js');
var defaultSettingsStore = require('./defaultSettingsStore-b9a20f38.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var R__default = /*#__PURE__*/_interopDefaultLegacy(R);
var AC__default = /*#__PURE__*/_interopDefaultLegacy(AC);
var T__default = /*#__PURE__*/_interopDefaultLegacy(T);

/**
 * Created by Andy Likuski on 2020.08.18
 * Copyright (c) 2020 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const {
  of
} = T__default['default'];
const tokenAuthOutputParams = {
  token: 1,
  payload: 1
};
const tokenAuthReadInputTypeMapper = {};
/**
 * Look for the auth token in the cache so we know if we can get an authenticated client
 * This doesn't work with a cache query so I'm currently using a fragement read
 * TokenAuth is a singleton in the cache so it doesn't need any props
 * @param {Object} apolloConfig
 * @param {Object} apolloConfig.apolloClient Needed for both component and client queries
 * @param {Object} props No props needed
 * @returns {*}
 */

const queryLocalTokenAuthContainer = (apolloConfig, props) => {
  // Unfortunately a cache miss throws
  try {
    return R__default['default'].compose( // Wrap in a task when we are doing apolloClient queries, otherwise we already have
    // a proper apollo container
    containerOrValue => R__default['default'].when(() => R__default['default'].propOr(false, 'apolloClient', apolloConfig), of)(containerOrValue), props => {
      return queryCacheHelpers.makeReadFragmentFromCacheContainer(apolloConfig, {
        name: 'tokenAuth',
        readInputTypeMapper: tokenAuthReadInputTypeMapper,
        outputParams: tokenAuthOutputParams
      }, // Pass all the props including the render function. Only __typenmae and id are needed by the fragment read
      R__default['default'].merge(props, // Singleton so id is just the type
      {
        __typename: 'ObtainJSONWebToken',
        id: 'ObtainJSONWebToken'
      }));
    })(props);
  } catch (e) {
    return containerHelpers.containerForApolloType(apolloConfig, {
      render: apolloClientCache.getRenderPropFunction(props),
      response: null
    });
  }
};
/**
 * Verifies an apolloClient auth token.
 * @param {Object} apolloClient
 * @param {Object} outputParams
 * @param {Object} props The props are optional. They are typically sent via the mutation
 * function in {variables: {username, password}}
 * @param {String} [props.username]
 * @param {String} [props.password]
 * @return {TasK|Object} Task or Apollo Component resolving to
 * {
  "data": {
    "tokenAuthMutation": {
      "token": the token
      "payload": {
        "username": the username
        "exp": Expiration time as 1598264561,
        "origIat": Original login time 1598264261
      },
    }
  }
}
 */

const tokenAuthMutationContainer = R__default['default'].curry((apolloConfig, {
  outputParams = tokenAuthOutputParams
}, props) => {
  return mutationHelpers.makeMutationRequestContainer(R__default['default'].merge(apolloConfig, {
    options: {
      variables: props => {
        return R__default['default'].pick(['username', 'password'], props);
      },
      update: (store, response) => {
        const tokenAuth = rescapeRamda.reqStrPathThrowing('data.tokenAuth', response); // This is what the Apollo Client reads to be authenticated

        localStorage.setItem('token', rescapeRamda.reqStrPathThrowing('token', tokenAuth)); // TODO Don't know if we need this in the cache
        // Mutate the cache with a singleton tokenAuth since we don't query for the tokenAuth

        mutationCacheHelpers.makeCacheMutation( // Use the store for writing if we don't have an apolloClient
        R__default['default'].merge({
          store
        }, apolloConfig), {
          name: 'tokenAuth',
          // output for the read fragment
          outputParams,
          // Write without @client fields
          force: true,
          singleton: true
        }, tokenAuth);
      }
    }
  }), {
    // Use this instead of name so that 'create' is prepended
    mutationNameOverride: 'tokenAuth',
    outputParams: outputParams || tokenAuthOutputParams,
    // Use username and password as flat variables, not as an ObtainJSONWebToken input type
    flattenVariables: true
  }, // Defaults are used to tell makeMutationRequestContainer about the expected variable types
  R__default['default'].merge({
    username: '',
    password: ''
  }, props));
});
/**
 * Deletes the token cookie of the current user
 * @param {Object} apolloClient
 * @param {Object} [outputParams] Defaults to {deleted: true}
 * @param {Object} props Should always be empty
 * @return {TasK|Object} Task or Apollo Component resolving to
 */

const deleteTokenCookieMutationRequestContainer = R__default['default'].curry((apolloConfig, {
  outputParams = null
}, props) => {
  return mutationHelpers.makeMutationRequestContainer(apolloConfig, {
    outputParams: outputParams || {
      deleted: 1
    },
    flattenVariables: true,
    mutationNameOverride: 'deleteTokenCookie'
  }, props);
});
/**
 * Deletes the token cookie for all users
 * @param {Object} apolloClient
 * @param {Object} [outputParams] Defaults to {deleted: true}
 * @param {Object} props Should always be empty
 * @return {TasK|Object} Task or Apollo Component resolving to
 */

const deleteRefreshTokenCookieMutationRequestContainer = R__default['default'].curry((apolloConfig, {
  outputParams = null
}, props) => {
  return mutationHelpers.makeMutationRequestContainer(apolloConfig, {
    outputParams: outputParams || {
      deleted: 1
    },
    flattenVariables: true,
    mutationNameOverride: 'deleteRefreshTokenCookie'
  }, props);
});
/**
 * Verifies an apolloClient auth token.
 * @param {Object} apolloClient
 * @param {Object} outputParams
 * @param {Object} props
 * @param {String} props.token The token to verify
 * @return {Function} Unary function expecting props and returning an Apollo Componnet or Task that resolves to the
 * token verification
 */

const verifyTokenMutationRequestContainer = R__default['default'].curry((apolloConfig, {
  outputParams = null
}, props) => {
  return mutationHelpers.makeMutationRequestContainer(apolloConfig, {
    outputParams: outputParams || {
      payload: 1
    },
    flattenVariables: true,
    mutationNameOverride: 'verifyToken'
  }, props);
});
/**
 * Refresh an apolloClient auth token.
 * @param {Object} apolloClient
 * @param {Object} outputParams
 * @param {Object} variables
 * @param {String} variables.token The token to verify
 * @return {Object} Task that resolves to the username, expiration (exp), and origlat (?)
 */

const refreshTokenMutationRequestContainer = R__default['default'].curry((apolloConfig, {
  outputParams = null
}, props) => {
  return mutationHelpers.makeMutationRequestContainer(apolloConfig, {
    outputParams: outputParams || {
      payload: 1
    },
    flattenVariables: true,
    mutationNameOverride: 'refreshToken'
  }, props);
});

/**
 * Created by Andy Likuski on 2020.08.18
 * Copyright (c) 2020 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const {
  of: of$1
} = T__default['default'];
const log = rescapeLog.loggers.get('rescapeDefault');
/**
 * Given a token returns a GraphQL client
 * @param {Object} config
 * @param {String} config.uri Graphpl URL, e.g.  'http://localhost:8000/api/graphql';
 * @param {Object} config.stateLinkResolvers: Resolvers for the stateLink, meaning local caching
 * local storage to store are auth token
 * @param {Function} config.writeDefaults expecting apolloClient that writes defaults ot the cache
 * @param {Object} config.settingsConfig
 * @param {Array|Object} config.settingsConfig.settingsOutputParams The settings outputParams
 * @param {[String]} config.settingsConfig.cacheOnlyObjs See defaultSettingsStore for an example
 * @param {[String]} config.settingsConfig.cacheIdProps See defaultSettingsStore for an example
 * @param {[String]} config.cacheData Existing client cache data if going from unauthorized to authorized
 * @param {String} authToken: If non-null, authenticates the client. Otherwise a non-auth Apollo Client
 * is created
 * @return {{apolloClient: ApolloClient}}
 */

const getOrCreateApolloClientTaskAndSetDefaults = ({
  // Existing client cache data if going from unauthorized to authorized
  cacheData,
  cacheOptions,
  uri,
  stateLinkResolvers,
  writeDefaults,
  settingsConfig
}, authToken) => {
  const {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  } = settingsConfig;
  return rescapeRamda.composeWithChain([({
    apolloConfig,
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams,
    writeDefaults
  }) => {
    const apolloClient = rescapeRamda.reqStrPathThrowing('apolloClient', apolloConfig); // Set writeDefaults to reset the cache. reset: true tells the function that this isn't the initial call

    apolloClient.onResetStore(() => writeDefaults(apolloClient, {
      cacheOnlyObjs,
      cacheIdProps,
      settingsOutputParams
    }));

    const taskIfNotTask = obj => {
      return R__default['default'].unless(obj => 'run' in obj, of$1)(obj);
    }; // Write the initial defaults as a task if not one already, return the apolloClient
    // The initial write sets reset: false in case we need to go to the server the first time to get the values
    // or the structure of the values


    return R__default['default'].map(() => ({
      apolloClient
    }), taskIfNotTask(writeDefaults(apolloClient, {
      cacheOnlyObjs,
      cacheIdProps,
      settingsOutputParams
    })));
  }, rescapeRamda.mapToNamedResponseAndInputs('user', ({
    apolloConfig,
    authToken
  }) => {
    // Fetch the user to get it into the cache so we know we are authenticated
    return R__default['default'].ifElse(R__default['default'].identity, () => defaultSettingsStore.currentUserQueryContainer(apolloConfig, defaultSettingsStore.userOutputParams, {}), () => of$1(null))(authToken);
  }), rescapeRamda.mapToNamedResponseAndInputs('apolloConfig', ({
    cacheData,
    uri,
    stateLinkResolvers,
    authToken
  }) => {
    // Memoized call
    return apolloClient.getOrCreateApolloClientTask({
      // Existing apolloClient to add auth to
      // If we have an unauthorized client we want to authorize it so we can maintain its cache
      cacheData,
      cacheOptions,
      uri,
      stateLinkResolvers,
      fixedHeaders: {
        authorization: authToken ? `JWT ${authToken}` : ''
      }
    });
  })])({
    cacheData,
    uri,
    stateLinkResolvers,
    authToken,
    writeDefaults,
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  });
};
/**
 * Given a userLogin with a tokenAuthMutation.token create the getApolloClientTask and return it and the token
 * This method is synchronous but returns a Task to be used in API chains
 * @param {Object} config
 * @param {Object} [config.cacheData] Existing cache data from a no auth apolloClient
 * @param {Object} config.cacheOptions
 * @param {String} config.uri Graphpl URL, e.g.  'http://localhost:8000/api/graphql';
 * @param {Object} config.stateLinkResolvers Resolvers for the stateLink, meaning local caching
 * @param {Function} config.writeDefaults
 * @param {Array|Object} config.outputParams Teh settings outputParams
 * @param {Object} token Return value from loginMutationTask() api call
 * @param {Array|Object} config.settingsConfig.defaultSettingsOutputParams The settings outputParams
 * @param {[String]} config.settingsConfig.defaultSettingsCacheOnlyObjs See defaultSettingsStore for an example
 * @param {[String]} config.settingsConfig.defaultSettingsCacheIdProps See defaultSettingsStore for an example
 * @return {Task<Object>} Task resolving to an object containing and object with a apolloClient, token.
 */

const getOrCreateAuthApolloClientWithTokenTask = R__default['default'].curry(({
  cacheData,
  cacheOptions,
  uri,
  stateLinkResolvers,
  writeDefaults,
  settingsConfig
}, token) => {
  const {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  } = settingsConfig;
  return R__default['default'].map(obj => {
    return R__default['default'].merge({
      token
    }, obj);
  }, getOrCreateApolloClientTaskAndSetDefaults({
    cacheData,
    cacheOptions,
    uri,
    stateLinkResolvers,
    writeDefaults,
    settingsConfig: {
      cacheOnlyObjs,
      cacheIdProps,
      settingsOutputParams
    }
  }, token));
});
/**
 * Given a userLogin with a tokenAuth.token create the getApolloClientTask and return it and the token
 * This method is synchronous but returns a Task to be used in API chains
 * @param {Object} config
 * @param {Object} config.cacheOptions
 * @param {String} config.uri Graphpl URL, e.g.  'http://localhost:8000/api/graphql';
 * @param {Object} config.stateLinkResolvers Resolvers for the stateLink, meaning local caching
 * @param {Function} config.writeDefaults
 * @param {Array|Object} config.outputParams Teh settings outputParams
 * @param {Object} config.settingsConfig
 * @param {Array|Object} config.settingsConfig.defaultSettingsOutputParams The settings outputParams
 * @param {[String]} config.settingsConfig.defaultSettingsCacheOnlyObjs See defaultSettingsStore for an example
 * @param {[String]} config.settingsConfig.defaultSettingsCacheIdProps See defaultSettingsStore for an example
 * @return {Task<Object>} Task resolving to an object containing and object with a no auth apolloClient and null token.
 */

const getOrCreateNoAuthApolloClientTask = R__default['default'].curry(({
  cacheOptions,
  uri,
  stateLinkResolvers,
  writeDefaults,
  settingsConfig
}) => {
  const {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  } = settingsConfig;
  return R__default['default'].map(obj => {
    // Matches resolved value of getOrCreateAuthApolloClientWithTokenTask
    return R__default['default'].merge({
      token: null
    }, obj);
  }, getOrCreateApolloClientTaskAndSetDefaults({
    cacheOptions,
    uri,
    stateLinkResolvers,
    writeDefaults,
    settingsConfig: {
      cacheOnlyObjs,
      cacheIdProps,
      settingsOutputParams
    }
  }, // null authToken
  null));
});

/**
 * Created by Andy Likuski on 2018.04.25
 * Copyright (c) 2018 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const {
  of: of$2
} = T__default['default'];
const {
  ApolloClient
} = AC__default['default'];
/**
 * Login and return an authenticated client task
 * @param {Object} config
 * @param {Object} config.cacheOptions
 * @param {String} config.uri graphql uri
 * @param {Object} config.stateLinkResolvers: Resolvers for the stateLink, meaning local caching
 * @param {Object} props
 * @param {String} props.username The username
 * @param {String} props.password The password
 * @param {Object} config.settingsConfig
 * @param {Array|Object} config.settingsConfig.defaultSettingsOutputParams The settings outputParams
 * @param {[String]} config.settingsConfig.defaultSettingsCacheOnlyObjs See defaultSettingsStore for an example
 * @param {[String]} config.settingsConfig.defaultSettingsCacheIdProps See defaultSettingsStore for an example
 * @return {{apolloClient: ApolloClient, token}}
 */

const loginToAuthClientTask = R__default['default'].curry(({
  cacheOptions,
  uri,
  stateLinkResolvers,
  writeDefaults,
  settingsConfig: {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  }
}, props) => {
  return rescapeRamda.composeWithChain([// loginResult.data contains {tokenAuth: token}
  // TODO can we modify noAuthApolloClientTask by writing the auth data to the cache instead??
  ({
    apolloConfig,
    user,
    loginData
  }) => {
    return of$2(R__default['default'].mergeAll([apolloConfig, {
      user: rescapeRamda.reqStrPathThrowing('data.currentUser', user)
    }, R__default['default'].pick(['token', 'payload'], rescapeRamda.reqStrPathThrowing('tokenAuth', loginData))]));
  }, rescapeRamda.mapToNamedResponseAndInputs('user', ({
    apolloConfig
  }) => {
    return defaultSettingsStore.currentUserQueryContainer(apolloConfig, defaultSettingsStore.userOutputParams, {});
  }), rescapeRamda.mapToNamedPathAndInputs('loginData', 'data', ({
    apolloConfig,
    props
  }) => {
    return tokenAuthMutationContainer(apolloConfig, {
      outputParams: tokenAuthOutputParams
    }, props);
  }), rescapeRamda.mapToNamedResponseAndInputs('apolloConfig', ({
    uri,
    stateLinkResolvers
  }) => {
    // Use unauthenticated ApolloClient for login
    return apolloClient.noAuthApolloClientTask({
      cacheOptions,
      uri,
      stateLinkResolvers
    });
  })])({
    uri,
    stateLinkResolvers,
    props
  });
});
/**
 * Return an nauthenticated client task
 * @param {Object} config
 * @param {Object} config.cacheOptions
 * @param {String} config.uri graphql uri
 * @param {Object} config.stateLinkResolvers: Resolvers for the stateLink, meaning local caching
 * @param {Object} config.settingsConfig
 * @param {Array|Object} config.settingsConfig.defaultSettingsOutputParams The settings outputParams
 * @param {[String]} config.settingsConfig.defaultSettingsCacheOnlyObjs See defaultSettingsStore for an example
 * @param {[String]} config.settingsConfig.defaultSettingsCacheIdProps See defaultSettingsStore for an example
 * @return {{apolloClient: ApolloClient, token}}
 */

const noLoginToAuthClientTask = R__default['default'].curry(({
  cacheOptions,
  uri,
  stateLinkResolvers,
  writeDefaults,
  settingsConfig: {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  }
}) => {
  return getOrCreateNoAuthApolloClientTask({
    cacheOptions,
    uri,
    stateLinkResolvers,
    writeDefaults,
    settingsConfig: {
      cacheOnlyObjs,
      cacheIdProps,
      settingsOutputParams
    }
  });
});
/**
 * Expects a GraphQLClient if already authenticated or login data if not
 * @param {Object} config
 * @param {Object} config.cacheOptions
 * @param {String} config.uri The URL to create client with if authentication is not already a GraphQLClient
 * @param {Object} config.stateLinkResolvers: Resolvers for the stateLink, meaning local caching
 * @param {Array} config.defaultSettingsOutputParams the settings output params
 * @param {Function} config.writeDefaults. Function to write default values to the client
 * @param {Object} config.settingsConfig
 * @param {Array|Object} config.settingsConfig.defaultSettingsOutputParams The settings outputParams
 * @param {[String]} config.settingsConfig.defaultSettingsCacheOnlyObjs See defaultSettingsStore for an example
 * @param {[String]} config.settingsConfig.defaultSettingsCacheIdProps See defaultSettingsStore for an example
 * @param {GraphQLClient|Object} authentication. If a GraphQLClient, a client with authentication already
 * in the header, such as an auth token. If an object, then username and password
 * @returns {Task<Object>} {apolloClient: Authorized Apollo Client, token: The authentication token,
 * function to clear the link state}
 */

const authClientOrLoginTask = R__default['default'].curry(({
  cacheOptions,
  uri,
  stateLinkResolvers,
  writeDefaults,
  settingsConfig: {
    cacheOnlyObjs,
    cacheIdProps,
    settingsOutputParams
  }
}, authentication) => {
  return R__default['default'].ifElse(({
    authentication
  }) => R__default['default'].is(ApolloClient, authentication), // Just wrap it in a task to match the other option
  apolloClient => of$2({
    apolloClient
  }), rescapeRamda.composeWithChain([// map userLogin to getApolloClientTask and token
  ({
    apolloConfig,
    uri,
    stateLinkResolvers,
    loginAuthentication
  }) => {
    return getOrCreateAuthApolloClientWithTokenTask({
      cacheData: rescapeRamda.reqStrPathThrowing('apolloClient.cache.data.data', apolloConfig),
      cacheOptions,
      uri,
      stateLinkResolvers,
      writeDefaults,
      settingsConfig: {
        cacheOnlyObjs,
        cacheIdProps,
        settingsOutputParams,
        defaultSettingsTypenames: defaultSettingsStore.defaultSettingsTypenames
      }
    }, R__default['default'].prop('data', loginAuthentication));
  }, rescapeRamda.mapToNamedResponseAndInputs('loginAuthentication', ({
    apolloConfig,
    authentication
  }) => {
    return tokenAuthMutationContainer(apolloConfig, {}, authentication);
  }), // map login values to token
  rescapeRamda.mapToNamedResponseAndInputs('apolloConfig', ({
    uri,
    stateLinkResolvers
  }) => {
    return apolloClient.noAuthApolloClientTask({
      cacheOptions,
      uri,
      stateLinkResolvers
    });
  })]))({
    uri,
    stateLinkResolvers,
    authentication: authentication
  });
});

exports.authClientOrLoginTask = authClientOrLoginTask;
exports.deleteRefreshTokenCookieMutationRequestContainer = deleteRefreshTokenCookieMutationRequestContainer;
exports.deleteTokenCookieMutationRequestContainer = deleteTokenCookieMutationRequestContainer;
exports.getOrCreateApolloClientTaskAndSetDefaults = getOrCreateApolloClientTaskAndSetDefaults;
exports.getOrCreateAuthApolloClientWithTokenTask = getOrCreateAuthApolloClientWithTokenTask;
exports.getOrCreateNoAuthApolloClientTask = getOrCreateNoAuthApolloClientTask;
exports.loginToAuthClientTask = loginToAuthClientTask;
exports.noLoginToAuthClientTask = noLoginToAuthClientTask;
exports.queryLocalTokenAuthContainer = queryLocalTokenAuthContainer;
exports.refreshTokenMutationRequestContainer = refreshTokenMutationRequestContainer;
exports.tokenAuthMutationContainer = tokenAuthMutationContainer;
exports.verifyTokenMutationRequestContainer = verifyTokenMutationRequestContainer;
//# sourceMappingURL=login-a898b78e.js.map
