'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var R = require('ramda');
var util = require('util');
require('ramda-lens');
require('pluralize');
var rescapeRamda = require('rescape-ramda');
require('folktale/result');
var requestHelpers = require('./requestHelpers.js');
var AC = require('@apollo/client');
require('@apollo/link-context');
require('@apollo/link-error');
var T = require('folktale/concurrency/task');
require('folktale/maybe');
require('react-apollo');
require('rescape-helpers-component');
var graphql = require('graphql');
require('node-fetch');
var rescapeLog = require('rescape-log');
require('apollo-cache-persist');
var rescapeValidate = require('rescape-validate');
var PropTypes = require('prop-types');
var apolloClient = require('./apolloClient.js');
var inflected = require('inflected');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var R__default = /*#__PURE__*/_interopDefaultLegacy(R);
var AC__default = /*#__PURE__*/_interopDefaultLegacy(AC);
var T__default = /*#__PURE__*/_interopDefaultLegacy(T);
var PropTypes__default = /*#__PURE__*/_interopDefaultLegacy(PropTypes);

/**
 * Created by Andy Likuski on 2018.05.10
 * Copyright (c) 2018 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
const {
  gql
} = AC__default['default'];
const {
  of
} = T__default['default'];
const log = rescapeLog.loggers.get('rescapeDefault');
/**
 * Makes a graphql query based on the queryParams
 * @param {String} queryName
 * @param {Object} inputParamTypeMapper maps Object params paths to the correct input type for the query
 * e.g. { 'data': 'DataTypeRelatedReadInputType' }
 * @param {Array|Object} outputParams
 * @param {Object} queryArguments
 * @returns {String} The query in a string
 */

const makeQuery = R__default['default'].curry((queryName, inputParamTypeMapper, outputParams, queryArguments) => {
  return _makeQuery({}, queryName, inputParamTypeMapper, outputParams, queryArguments);
});
/**
 * Creates a fragment query for fetching values from the cache
 * @param {String} queryName Unique query name
 * @param {Object} inputParamTypeMapper Used to generate the correct complex input types
 * @param {Array|Object} outputParams
 * @param props Only for __typename
 * @param {String} [props.__typename] Only required for fragment queries
 * I think fragments never need args so only queryArguments.__typename should be specified for fragment queries
 * @return {string} The query string, not gql
 * @type {any}
 */

const makeFragmentQuery = R__default['default'].curry((queryName, inputParamTypeMapper, outputParams, props) => {
  return _makeQuery({
    isFragment: true
  }, queryName, inputParamTypeMapper, outputParams, props);
});
/***
 *
 * Makes a query or fragment query.
 * Memoized so we don't give Apollo the same query twice
 * @param {Object} queryConfig
 * @param {Boolean} queryConfig.client If true adds a client directive
 * @param {Boolean} queryConfig.isFragment If true creates a fragment
 * @param queryName
 * @param inputParamTypeMapper
 * @param {Array|Object} outputParams
 * @param props
 * @param {String} [props.__typename] Only required for fragment queries
 * I think fragments never need args so only queryArguments.__typename should be specified for fragment queries
 * @return {string} The query string, not gql
 * @private
 */

const _makeQuery = rescapeRamda.memoized((queryConfig, queryName, inputParamTypeMapper, outputParams, props) => {
  const resolve = requestHelpers.resolveGraphQLType(inputParamTypeMapper); // Never allow __typename. It might be in the queryArguments if the they come from the output of another query

  const cleanedQueryArguments = rescapeRamda.omitDeep(['__typename'], props); // These are the first line parameter definitions of the query, which list the name and type

  const params = R__default['default'].join(', ', rescapeRamda.mapObjToValues((value, key) => {
    // Map the key to the inputParamTypeMapper value for that key if given
    // This is only needed when value is an Object since it needs to map to a custom graphql inputtype
    return `$${key}: ${resolve(key, value)}!`;
  }, cleanedQueryArguments));

  const parenWrapIfNotEmpty = str => R__default['default'].unless(R__default['default'].isEmpty, str => `(${str})`, str); // These are the second line arguments that map parameters to variables


  const args = R__default['default'].join(', ', rescapeRamda.mapObjToValues((value, key) => {
    return `${key}: $${key}`;
  }, cleanedQueryArguments)); // Only use parens if there are actually variables/arguments

  const variableString = R__default['default'].ifElse(R__default['default'].length, R__default['default'].always(params), R__default['default'].always(''))(R__default['default'].keys(cleanedQueryArguments));
  const clientTokenIfClientQuery = R__default['default'].ifElse(R__default['default'].prop('client'), R__default['default'].always('@client'), R__default['default'].always(null))(queryConfig); // Either we have a query queryName or fragment queryName on queryArguments.__typename
  // I think fragments never need args so only queryArguments.__typename should be specified for fragment queryies

  const queryOrFragment = R__default['default'].ifElse(R__default['default'].prop('isFragment'), R__default['default'].always(`fragment ${queryName}Fragment on ${R__default['default'].prop('__typename', props)}`), R__default['default'].always(`query ${queryName}`))(queryConfig); // Unless we are creating a fragment, wrap the outputParams in the name of the type we are querying

  const unlessFragment = content => R__default['default'].ifElse(R__default['default'].prop('isFragment'), R__default['default'].always(null), R__default['default'].always(content))(queryConfig);

  const output = R__default['default'].join('', rescapeRamda.compact([unlessFragment(R__default['default'].join(' ', rescapeRamda.compact([queryName, parenWrapIfNotEmpty(args), clientTokenIfClientQuery, '{']))), requestHelpers.formatOutputParams(outputParams), unlessFragment('}')])); // We use the queryName as the label of the query and the name that matches the schema

  return `${queryOrFragment} ${parenWrapIfNotEmpty(variableString)} { 
  ${output}
}`;
});
/**
 * Creates a query task for any type
 * @param {Object} apolloConfig The Apollo configuration with either an ApolloClient for server work or an
 * Apollo wrapped Component for browser work
 * @param {Object} apolloConfig.apolloClient Optional Apollo client, authenticated for most calls
 * @param {Object} apolloConfig.apolloComponent Optional Apollo component
 * @param {Function} apolloConfig.apolloComponent.options Required for ApolloComponent container queries.
 * A unary function expecting components from the parent component or container
 * @param {Object} apolloConfig.apolloComponent.options.variables Variables for the ApolloComponent container
 * @param {Object} apolloConfig.apolloComponent.options.errorPolicy Optional errorPolicy string for the ApolloComponent
 * container
 * @param {String} name The lowercase name of the object matching the query name, e.g. 'regions' for regionsQuery
 * @param {Object} readInputTypeMapper maps object keys to complex input types from the Apollo schema. Hopefully this
 * will be automatically resolved soon. E.g. {data: 'DataTypeofLocationTypeRelatedReadInputType'}
 * @param {Function} [normalizeProps] Optional function to normalize props after calling
 * apolloConfig.options.variables if defined.
 * @param {Array|Object} outputParams output parameters for the query in this style json format:
 *  [
 *    'id',
 *    {
 *        data: [
 *         'foo',
 *         {
 *            properties: [
 *             'type',
 *            ]
 *         },
 *         'bar',
 *       ]
 *    }
 *  ]
 *
 *  In other words, start every type as a list and embed object types using {objectTypeKey: [...]}
 *  props ahead of time with the container. It should be the expected prop names and
 *  example value types (e.g. 0 for Number) TODO we could use types instead of numbers, if we can figure out a type
 *  to identify primitives
 *  @param {Object} props. The props for the query or an Apollo container that will supply the props
 *  @param {Task} An apollo query task that resolves to and object with the results of the query. Successful results
 *  are in obj.data[name]. Errors are in obj.errors. Since the queries are stored in data[name], multiple queries
 *  of different could be merged together into the data field. This also matches what Apollo components expect.
 *  If you need the value in a Result.Ok or Result.Error to halt operations on error, use requestHelpers.mapQueryContainerToNamedResultAndInputs
 */

const makeQueryContainer = rescapeValidate.v(R__default['default'].curry((apolloConfig, {
  name,
  readInputTypeMapper,
  outputParams,
  normalizeProps
}, props) => {
  // Limits the arguments the query uses based on apolloConfig.options.variables(props) if specified
  const nameTitle = R__default['default'].compose(rescapeRamda.capitalize, inflected.singularize)(name);
  const readInputTypeMapperOrDefault = R__default['default'].defaultTo({
    'data': `${nameTitle}DataTypeof${nameTitle}TypeRelatedReadInputType`
  }, readInputTypeMapper); // See if the query is ready to run

  const skip = rescapeRamda.strPathOr(false, 'options.skip', apolloConfig); // If not skip, process the props

  const winnowedProps = R__default['default'].ifElse(() => skip, () => ({}), props => R__default['default'].compose( // Normalize the props if needed. This removes top-level key/values that might be in the object that
  // aren't expected/needed by the API
  props => {
    return (normalizeProps || R__default['default'].identity)(props);
  }, props => {
    // Use apolloConfig.options.variables function to filter if specified.
    // This extracts the needed props for the query
    return requestHelpers._winnowRequestProps(apolloConfig, props);
  })(props))(props);
  const query = gql`${makeQuery(name, readInputTypeMapperOrDefault, outputParams, winnowedProps)}`;
  log.debug(`Creating Query:\n${graphql.print(query)}\nArguments:\n${R__default['default'].ifElse(() => skip, () => 'Props are not ready', winnowedProps => util.formatWithOptions({
    depth: 10
  }, '%j', winnowedProps))(winnowedProps)}\n`);
  const componentOrTask = apolloClient.authApolloQueryContainer( // Send the normalizeProps function if defined as composed into function apolloConfig.options.variables
  R__default['default'].when(() => normalizeProps, apolloConfig => composePropsFilterIntoApolloConfigOptionsVariables(apolloConfig, normalizeProps))(apolloConfig), query, // Non-winnowed props because the component does calls its options.variables function
  props);
  return R__default['default'].when(componentOrTask => 'run' in componentOrTask, // If it's a task report the result. Components have to run their query
  componentOrTask => {
    return R__default['default'].map(queryResponse => {
      log.debug(`makeQueryTask for ${name} responded: ${rescapeRamda.replaceValuesWithCountAtDepthAndStringify(2, queryResponse)}`);
      return queryResponse;
    }, componentOrTask);
  })(componentOrTask);
}), [['apolloConfig', PropTypes__default['default'].shape().isRequired], ['queryOptions', PropTypes__default['default'].shape({
  name: PropTypes__default['default'].string.isRequired,
  readInputTypeMapper: PropTypes__default['default'].shape(),
  outputParams: PropTypes__default['default'].oneOfType([PropTypes__default['default'].array, PropTypes__default['default'].shape()]).isRequired
})], ['props', PropTypes__default['default'].shape().isRequired]], 'makeQueryContainer');
/**
 *
 * Use given props to call the function at requests.arg.options, and then get the .variables of the returned value
 * @param {Function} apolloComponent Expects props and returns an Apollo Component
 * @param {Object} props
 * @returns {Object} variables to use for the query
 */

const createRequestVariables = (apolloComponent, props) => {
  return rescapeRamda.reqStrPathThrowing('props.variables', apolloComponent(props));
};
/**
 * Runs the apollo queries in queryComponents as tasks if runContainerQueries is true. If true,
 * resolvedPropsTask are resolved and set to the queries. props are also returned independent of the queries
 * If runContainerQueries is false, just resolves resolvedPropsTask and return the props
 * @param {Task} resolvedPropsTask A task that resolves the props to use
 * @param {Object} queryTasks Keyed by name, valued by a queryTask that expects the props.
 * Each queryTask resolves to a response. Responses are combined and keyed by the name
 * The responses are combined
 * @param {boolean} [runContainerQueries] Default true. When true run the container queries
 * @return {Task} A task that resolves to the props of resolvedPropsTask merged with the query results if there
 * are any queries and runContainerQueries is true
 */

const apolloQueryResponsesTask = (resolvedPropsTask, queryTasks, runContainerQueries = true) => {
  // Task Object -> Task
  return rescapeRamda.composeWithChain([// Wait for all the queries to finish
  props => {
    const queryTasksOrNone = runContainerQueries && queryTasks ? queryTasks : {}; // Each query resolves and the values are assigned to the key and merged with the props
    // This is similar to how react-adopt calls our Apollo request components

    return rescapeRamda.composeWithChain([...R__default['default'].reverse(rescapeRamda.mapObjToValues((tsk, key) => {
      return rescapeRamda.mapToNamedResponseAndInputs(key, props => tsk(props));
    }, queryTasksOrNone)), // This is our returned task if there are no queries
    props => of(props)])(props);
  }, // Resolve the props from the task
  resolvedPropsTask => resolvedPropsTask.map(x => x)])(resolvedPropsTask);
};
/**
 * Modifies apolloConfig's options.variables function to additionally filter props
 * with propsFilter. options.variables may or may not already be defined as a filter
 * @param {Object} apolloConfig The apolloConfig
 * @param {Function} propsFilter Expects props and returns filtered props
 * @returns {*}
 */

const composePropsFilterIntoApolloConfigOptionsVariables = (apolloConfig, propsFilter) => {
  return R__default['default'].over(R__default['default'].lensPath(['options', 'variables']), variables => {
    return props => R__default['default'].compose(props => propsFilter(props), props => R__default['default'].when(() => R__default['default'].is(Function, variables), variables)(props))(props);
  }, apolloConfig);
};

exports._makeQuery = _makeQuery;
exports.apolloQueryResponsesTask = apolloQueryResponsesTask;
exports.composePropsFilterIntoApolloConfigOptionsVariables = composePropsFilterIntoApolloConfigOptionsVariables;
exports.createRequestVariables = createRequestVariables;
exports.makeFragmentQuery = makeFragmentQuery;
exports.makeQuery = makeQuery;
exports.makeQueryContainer = makeQueryContainer;
//# sourceMappingURL=queryHelpers.js.map
